#include "Random.hpp"

#include <cstring>
using namespace std;

#ifdef _WIN32
# define _WIN32_LEAN_AND_MEAN
# include <windows.h>
# include <wincrypt.h>
# pragma comment(lib, "advapi32.lib")
#else
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <unistd.h>
#endif


Random::Random()
{
	init();
}


static const u32 MEXP = 19937;
static const u32 N128 = MEXP/128 + 1;
static const u32 N64 = N128 * 2;
static const u32 N32 = N128 * 4;
static const u32 POS1 = 122;
static const u32 SL1 = 18;
static const u32 SL2 = 1;
static const u32 SL2BITS = SL2*8;
static const u32 SR1 = 11;
static const u32 SR2 = 1;
static const u32 SR2BITS = SR2*8;
static const u32 MSK1 = 0xdfffffefU;
static const u32 MSK2 = 0xddfecb7fU;
static const u32 MSK3 = 0xbffaffffU;
static const u32 MSK4 = 0xbffffff6U;
static const u32 PARITY[4] = {0x00000001U, 0x00000000U, 0x00000000U, 0x13c9e684U};


MersenneTwister::MersenneTwister()
{
	state32 = &state[0].u[0];
}

// fix the initial state to ensure that the full generator period will occur
void MersenneTwister::enforcePeriod()
{
	u32 inner = 0;

	// An odd number of parity bits set is OK
	inner ^= state32[0] & PARITY[0];
	inner ^= state32[1] & PARITY[1];
	inner ^= state32[2] & PARITY[2];
	inner ^= state32[3] & PARITY[3];
	inner ^= inner >> 16;
	inner ^= inner >> 8;
	inner ^= inner >> 4;
	inner ^= inner >> 2;
	inner ^= inner >> 1;
	if ((inner & 1)) return;

	// Otherwise, flip the lowest parity bit to make it odd
	for (u32 ii = 0; ii < 4; ++ii)
	{
		if (PARITY[ii])
		{
			state32[ii] ^= LEAST_SIGNIFICANT_BIT(PARITY[ii]);
			break;
		}
	}
}

// initialize the generator with a 32-bit seed
void MersenneTwister::init(u32 seed)
{
	state32[0] = seed;

	for (u32 ii = 1; ii < N32; ++ii)
		state32[ii] = 1812433253UL * (state32[ii-1] ^ (state32[ii-1] >> 30)) + ii;

	enforcePeriod();
	used = N32;
}

// initialize with an array of seeds
void MersenneTwister::init(u32 *seeds, u32 words)
{
	u32 ii, jj, r, count, mid, lag, size = N32;

	if (size >= 623)		lag = 11;
	else if (size >= 68)	lag = 7;
	else if (size >= 39)	lag = 5;
	else					lag = 3;

	mid = (size - lag) / 2;

	memset(state, 0x8b, sizeof(state));

	if (words+1 > N32)	count = words+1;
	else				count = N32;

#define FUNC1(x) ( ((x) ^ ((x) >> 27)) * (u32)1664525UL )
#define FUNC2(x) ( ((x) ^ ((x) >> 27)) * (u32)1566083941UL )

	r = FUNC1(state32[0]);
	state32[mid] += r;
	r += words;
	state32[mid + lag] += r;
	state32[0] = r;

	--count;
	for (ii = 1, jj = 0; (jj < count) && (jj < words); ++jj)
	{
		r = FUNC1(state32[ii] ^ state32[(ii + mid) % N32] ^ state32[(ii + N32 - 1) % N32]);
		state32[(ii + mid) % N32] += r;
		r += seeds[jj] + ii;
		state32[(ii + mid + lag) % N32] += r;
		state32[ii] = r;
		++ii;
		ii %= N32;
	}
	for (; jj < count; ++jj)
	{
		r = FUNC1(state32[ii] ^ state32[(ii + mid) % N32] ^ state32[(ii + N32 - 1) % N32]);
		state32[(ii + mid) % N32] += r;
		r += ii;
		state32[(ii + mid + lag) % N32] += r;
		state32[ii] = r;
		++ii;
		ii %= N32;
	}
	for (jj = 0; jj < N32; ++jj)
	{
		r = FUNC2(state32[ii] + state32[(ii + mid) % N32] + state32[(ii + N32 - 1) % N32]);
		state32[(ii + mid) % N32] ^= r;
		r -= ii;
		state32[(ii + mid + lag) % N32] ^= r;
		state32[ii] = r;
		++ii;
		ii %= N32;
	}

#undef FUNC1
#undef FUNC2

	enforcePeriod();
	used = N32;
}

void MersenneTwister::init()
{
	u32 seeds[16];

#ifdef _WIN32
	HCRYPTPROV hCryptProv;

	if (CryptAcquireContext(&hCryptProv, 0, 0, PROV_DSS, CRYPT_VERIFYCONTEXT))
	{
		CryptGenRandom(hCryptProv, sizeof(seeds), (BYTE*)seeds);
		CryptReleaseContext(hCryptProv, 0);
	}
#else
	int fd = open("/dev/urandom", O_RDONLY);
	if (fd >= 0)
	{
		read(fd, seeds, sizeof(seeds));
		close(fd);
	}
#endif // _WIN32

	init(seeds, 16);
}

// r != n, 0 < bits < 32
static inline void shiftLeft128(u128 *r, u128 *n, u32 bits)
{
	r->u[0] = n->u[0] << bits;
	r->u[1] = (n->u[1] << bits) | (n->u[0] >> (32 - bits)); 
	r->u[2] = (n->u[2] << bits) | (n->u[1] >> (32 - bits)); 
	r->u[3] = (n->u[3] << bits) | (n->u[2] >> (32 - bits)); 
}

// r != n, 0 < bits < 32
static inline void shiftRight128(u128 *r, u128 *n, u32 bits)
{
	r->u[0] = (n->u[0] >> bits) | (n->u[1] << (32 - bits));
	r->u[1] = (n->u[1] >> bits) | (n->u[2] << (32 - bits));
	r->u[2] = (n->u[2] >> bits) | (n->u[3] << (32 - bits));
	r->u[3] = n->u[3] >> bits; 
}

// a ^= (a << SL2BITS) ^ ((b >> SR1) & MSK) ^ (c >> SR2BITS) ^ (d{0..3} << SL1)
void MersenneTwister::round(u128 *a, u128 *b, u128 *c, u128 *d)
{
	u128 x, y;

	shiftLeft128(&x, a, SL2BITS);
	shiftRight128(&y, c, SR2BITS);

    a->u[0] ^= x.u[0] ^ ((b->u[0] >> SR1) & MSK1) ^ y.u[0] ^ (d->u[0] << SL1);
    a->u[1] ^= x.u[1] ^ ((b->u[1] >> SR1) & MSK2) ^ y.u[1] ^ (d->u[1] << SL1);
    a->u[2] ^= x.u[2] ^ ((b->u[2] >> SR1) & MSK3) ^ y.u[2] ^ (d->u[2] << SL1);
    a->u[3] ^= x.u[3] ^ ((b->u[3] >> SR1) & MSK4) ^ y.u[3] ^ (d->u[3] << SL1);
}

// permute the existing state into a new one
void MersenneTwister::update()
{
	u128 *r1 = state + N128 - 2;
	u128 *r2 = state + N128 - 1;
	u32 ii;

	for (ii = 0; ii < N128 - POS1; ++ii)
	{
		round(state + ii, state + ii + POS1, r1, r2);
		r1 = r2;
		r2 = state + ii;
	}

	for (; ii < N128; ++ii)
	{
		round(state + ii, state + ii + POS1 - N128, r1, r2);
		r1 = r2;
		r2 = state + ii;
	}

	used = 0;
}

// generate a 32-bit random number
u32 MersenneTwister::generate()
{
	if (used >= N32)
		update();

	return state32[used++];
}

// generate a series of random numbers to fill a buffer of any size
void MersenneTwister::generate(void *buffer, u32 bytes)
{
	u8 *buffer8 = (u8 *)buffer;
	u32 words = bytes / 4;

	while (words > 0)
	{
		if (used >= N32)
			update();

		u32 remaining = N32 - used;
		u32 copying = words;
		if (copying > remaining) copying = remaining;

		memcpy(buffer8, state32 + used, copying*4);
		used += copying;
		words -= copying;
		buffer8 += copying*4;
	}

	switch (bytes % 4)
	{
	case 1:
		buffer8[0] = (u8)generate();
		break;
	case 2:
		*(u16*)buffer8 = (u16)generate();
		break;
	case 3:
		words = generate();
		buffer8[0] = (u8)words;
		*(u16*)(buffer8+1) = (u16)(words >> 8);
		break;
	}
}


static const s32 STDLIB_RAND_MULTIPLIER = 214013;
static const s32 STDLIB_RAND_ADDEND = 2531011;
static const s32 STDLIB_RANDU_MULTIPLIER = 65539;

u16 StandardRand::rand()
{
	seed = STDLIB_RAND_MULTIPLIER * seed + STDLIB_RAND_ADDEND;
	return (u16)((seed >> 16) & 0x7fff);
}

u16 StandardRand::randu()
{
	seed *= STDLIB_RANDU_MULTIPLIER;
	return (u16)((seed >> 16) & 0x7fff);
}


u32 NLCRand32(int x, int y, u32 seed)
{
	// Map 2D->1D
	// To avoid repetitive output, adjust pitch constant (271) higher if needed
    u32 n = x + y * 271 + seed;

    // Mixing step
    n = (n << 13) ^ n;

    // Non-linear congruential mixing step
    n *= (n * n * 15731 + 789221) + 1376312589;

    return n;
}


float NLCRandNorm(int x, int y, u32 seed)
{
	u32 n = NLCRand32(x, y, seed);

    // Clamp to -1..1 (mode preferred by my Perlin noise code)
    return 1.0f - n / 2147483647.5f;
}
